from setuptools import setup


setup(
    name="pyqt5-tile-layout",
    version="0.1.3",
    author="Arnaud Frammery",
    author_email="arnaud.frammery@gmail.com",
    description="A tile layout for PyQt5",
    long_description="See the github project README for examples and documentation: "
                     "https://github.com/arnaudframmery/qt-tile-layout",
    license="GPL3",
    url="https://github.com/arnaudframmery/qt-tile-layout",
    packages=["QTileLayout"],
    install_requires=["PyQt5~=5.12"],
)
